"""
Status factory module.
"""
from __future__ import annotations

import typing

from digitalhub_core.entities._base.status import State
from digitalhub_core.utils.import_utils import (
    check_layer_existence,
    check_module_existence_by_framework,
    import_class,
    import_registry,
)

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.status import Status


def build_status(
    kind: str,
    layer_digitalhub: str | None = None,
    framework_runtime: str | None = None,
    **kwargs,
) -> Status:
    """
    Build entity status object.

    Parameters
    ----------
    kind : str
        The type of status to build.
    layer_digitalhub : str
        Layer name (e.g. "digitalhub_core") where the status is defined.
    framework_runtime : str
        Framework name (e.g. "mlrun", "nefertem") where the status is defined.
        This referes to the module where the status is defined. The import
        function will search in different layers.

    Returns
    -------
    Status
        Status object.
    """
    # Raise error if both layer and framework are not provided or if both are provided.
    # This is because we need some information where to search for the status class.
    if (layer_digitalhub is None) == (framework_runtime is None):
        raise ValueError("Either layer or framework must be provided, but not both.")

    # If layer is not provided, try to import from runtime, otherwise search by layer.
    if framework_runtime is not None:
        layer_digitalhub = check_module_existence_by_framework(framework_runtime)
    else:
        check_layer_existence(layer_digitalhub)

    # Import registry
    # Note that this requires the creation of a registry whenever a new status entity is created.
    module_to_import = f"{layer_digitalhub}.entities.registries"
    registry = import_registry(module_to_import, "status_registry")
    class_status = _import_class(registry, kind)

    kwargs = parse_arguments(**kwargs)
    return class_status(**kwargs)


def parse_arguments(**kwargs) -> dict:
    """
    Parse keyword arguments and add default values if necessary.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    dict
        Keyword arguments with default values.
    """
    state = kwargs.get("state")
    if state is None:
        kwargs["state"] = State.CREATED.value
    else:
        if kwargs["state"] not in State.__members__:
            raise ValueError(f"Invalid state: {state}")
    return kwargs


def _import_class(registry: dict, kind: str) -> Status:
    """
    Import status class from registry.

    Parameters
    ----------
    registry : dict
        Registry.
    kind : str
        Entity kind.

    Returns
    -------
    Status
        Status class.
    """
    try:
        module_and_class = registry[kind]
    except KeyError:
        raise ValueError(f"Status {kind} not found in registry.")

    try:
        module = module_and_class["module"]
        status_class = module_and_class["status_class"]
    except KeyError:
        raise ValueError("Malformed registration.")

    return import_class(module, status_class)



"""
Module for performing operations on tasks.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import check_context, get_context
from digitalhub_core.entities.tasks.entity import task_from_dict, task_from_parameters
from digitalhub_core.utils.api import api_ctx_delete, api_ctx_list, api_ctx_read, api_ctx_update
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.tasks.entity import Task


ENTITY_TYPE = "tasks"


def create_task(**kwargs) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    return task_from_parameters(**kwargs)


def create_task_from_dict(obj: dict) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    obj : dict
        Object dictionary.

    Returns
    -------
    Task
       Object instance.
    """
    check_context(obj.get("project"))
    return task_from_dict(obj)


def new_task(
    project: str,
    kind: str,
    uuid: str | None = None,
    source: str | None = None,
    labels: list[str] | None = None,
    function: str | None = "",
    node_selector: list[dict] | None = None,
    volumes: list[dict] | None = None,
    resources: list[dict] | None = None,
    affinity: dict | None = None,
    tolerations: list[dict] | None = None,
    k8s_labels: list[dict] | None = None,
    env: list[dict] | None = None,
    secrets: list[str] | None = None,
    backoff_limit: int | None = None,
    schedule: str | None = None,
    replicas: int | None = None,
    **kwargs,
) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    project : str
        Project name.
    kind : str
        Kind of the object.
    uuid : str
        ID of the object in form of UUID.
    source : str
        Remote git source for object.
    labels : list[str]
        List of labels.
    function : str
        The function string identifying the function.
    node_selector : list[NodeSelector]
        The node selector of the task.
    volumes : list[Volume]
        The volumes of the task.
    resources : list[Resource]
        Kubernetes resources for the task.
    affinity : Affinity
        The affinity of the task.
    tolerations : list[Toleration]
        The tolerations of the task.
    k8s_labels : list[Label]
        The labels of the task.
    env : list[Env]
        The env variables of the task.
    secrets : list[str]
        The secrets of the task.
    backoff_limit : int
        The backoff limit of the task.
    schedule : str
        The schedule of the task.
    replicas : int
        The replicas of the task.
    **kwargs
        Spec keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    obj = create_task(
        project=project,
        kind=kind,
        uuid=uuid,
        source=source,
        labels=labels,
        function=function,
        node_selector=node_selector,
        volumes=volumes,
        resources=resources,
        affinity=affinity,
        tolerations=tolerations,
        k8s_labels=k8s_labels,
        env=env,
        secrets=secrets,
        backoff_limit=backoff_limit,
        schedule=schedule,
        replicas=replicas,
        **kwargs,
    )
    obj.save()
    return obj


def get_task(project: str, entity_id: str, **kwargs) -> Task:
    """
    Get object from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Task
        Object instance.
    """
    api = api_ctx_read(project, ENTITY_TYPE, entity_id)
    obj = get_context(project).read_object(api, **kwargs)
    return create_task_from_dict(obj)


def import_task(file: str) -> Task:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Task
        Object instance.
    """
    obj: dict = read_yaml(file)
    return create_task_from_dict(obj)


def delete_task(project: str, entity_id: str, cascade: bool = True, **kwargs) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    params = kwargs.get("params", {})
    if params is None or not params:
        kwargs["params"] = {}
        kwargs["params"]["cascade"] = str(cascade).lower()
    api = api_ctx_delete(project, ENTITY_TYPE, entity_id)
    return get_context(project).delete_object(api, **kwargs)


def update_task(entity: Task, **kwargs) -> dict:
    """
    Update object in backend.

    Parameters
    ----------
    entity : Task
        The object to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_ctx_update(entity.project, ENTITY_TYPE, entity.id)
    return get_context(entity.project).update_object(api, entity.to_dict(), **kwargs)


def list_tasks(project: str, **kwargs) -> list[dict]:
    """
    List all objects from backend.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    list[dict]
        List of tasks dict representations.
    """
    api = api_ctx_list(project, ENTITY_TYPE)
    return get_context(project).list_objects(api, **kwargs)



"""
Function module.
"""
from __future__ import annotations

import typing
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.tasks.crud import create_task, create_task_from_dict, delete_task, new_task
from digitalhub_core.utils.api import api_ctx_create, api_ctx_list, api_ctx_update
from digitalhub_core.utils.exceptions import BackendError, EntityError
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.functions.metadata import FunctionMetadata
    from digitalhub_core.entities.functions.spec import FunctionSpec
    from digitalhub_core.entities.functions.status import FunctionStatus
    from digitalhub_core.entities.runs.entity import Run
    from digitalhub_core.entities.tasks.entity import Task


class Function(Entity):
    """
    A class representing a function.
    """

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: FunctionMetadata,
        spec: FunctionSpec,
        status: FunctionStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        project : str
            Project name.
        name : str
            Name of the object.
        uuid : str
            Version of the object.
        kind : str
            Kind of the object.
        metadata : FunctionMetadata
            Metadata of the object.
        spec : FunctionSpec
            Specification of the object.
        status : FunctionStatus
            Status of the object.
        """
        super().__init__()
        self.project = project
        self.name = name
        self.id = uuid
        self.kind = kind
        self.key = f"store://{project}/functions/{kind}/{name}:{uuid}"
        self.metadata = metadata
        self.spec = spec
        self.status = status

        # Add attributes to be used in the to_dict method
        self._obj_attr.extend(["project", "name", "id", "key"])

        # Initialize tasks
        self._tasks: dict[str, Task] = {}

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> Function:
        """
        Save entity into backend.

        Parameters
        ----------
        uuid : str
            Specify uuid for the function update

        Returns
        -------
        Function
            Entity saved.
        """
        obj = self.to_dict(include_all_non_private=True)

        if not update:
            api = api_ctx_create(self.project, "functions")
            new_obj = self._context().create_object(api, obj)
            self._update_attributes(new_obj)
            return self

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_ctx_update(self.project, "functions", self.id)
        new_obj = self._context().update_object(api, obj)
        self._update_attributes(new_obj)
        return self

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()

        if filename is None:
            filename = f"function_{self.kind}_{self.name}_{self.id}.yml"
        pth = Path(self._context().project_dir) / filename
        pth.parent.mkdir(parents=True, exist_ok=True)

        # Embed tasks in file
        if self._tasks:
            obj = [obj] + [v.to_dict() for _, v in self._tasks.items()]

        write_yaml(pth, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.project)

    #############################
    #  Function Methods
    #############################

    def run(
        self,
        action: str,
        node_selector: list[dict] | None = None,
        volumes: list[dict] | None = None,
        resources: list[dict] | None = None,
        affinity: dict | None = None,
        tolerations: list[dict] | None = None,
        labels: list[dict] | None = None,
        env: list[dict] | None = None,
        secrets: list[str] | None = None,
        backoff_limit: int | None = None,
        schedule: str | None = None,
        replicas: int | None = None,
        inputs: dict | None = None,
        outputs: dict | None = None,
        parameters: dict | None = None,
        values: list | None = None,
        local_execution: bool = False,
        **kwargs,
    ) -> Run:
        """
        Run function.

        Parameters
        ----------
        action : str
            Action to execute. Task parameter.
        node_selector : list[dict]
            The node selector of the task. Task parameter.
        volumes : list[dict]
            The volumes of the task. Task parameter.
        resources : list[dict]
            Kubernetes resources for the task. Task parameter.
        affinity : Affinity
            Kubernetes affinity parameters. Task parameter.
        tolerations : list[dict]
            Kubernetes tolerations parameters. Task parameter.
        labels : list[dict]
            The labels of the task.
        env : list[dict]
            The env variables of the task. Task parameter.
        secrets : list[str]
            The secrets of the task. Task parameter.
        backoff_limit : int
            The backoff limit of the task. Task parameter.
        schedule : str
            The schedule of the task. Task parameter.
        replicas : int
            The replicas of the task. Task parameter.
        inputs : dict
            Function inputs. Run parameter.
        outputs : dict
            Function outputs. Run parameter.
        parameters : dict
            Function parameters. Run parameter.
        values : list
            Function values. Run parameter.
        local_execution : bool
            Flag to determine if object has local execution. Run parameter.
        **kwargs
            Keyword arguments passed to Task builder.

        Returns
        -------
        Run
            Run instance.
        """

        # Create task if does not exists
        task = self._tasks.get(action)

        # Check in backend
        if task is None and not self._context().local:
            task = self._check_task_in_backend(action)

        # Create new task
        if task is None:
            task = self.new_task(
                kind=f"{self.kind}+{action}",
                node_selector=node_selector,
                volumes=volumes,
                resources=resources,
                affinity=affinity,
                tolerations=tolerations,
                labels=labels,
                env=env,
                secrets=secrets,
                backoff_limit=backoff_limit,
                schedule=schedule,
                replicas=replicas,
                **kwargs,
            )

        # Run function from task
        run = task.run(inputs, outputs, parameters, values, local_execution)

        # If execution is done by DHCore backend, return the object
        if not local_execution:
            if self._context().local:
                raise BackendError("Cannot run remote function with local backend.")
            return run

        # If local execution, build and launch run
        run.build()
        with ThreadPoolExecutor(max_workers=1) as executor:
            result = executor.submit(run.run)
        return result.result()

    def _check_task_in_backend(self, action: str) -> None | Task:
        """
        Check if task exists in backend.

        Parameters
        ----------
        action : str
            Action to check.

        Returns
        -------
        None | Task
            Task if exists, None otherwise.
        """
        api = api_ctx_list(self.project, "tasks")
        params = {"function": self._get_function_string(), "kind": f"{self.kind}+{action}"}
        objs = self._context().list_objects(api, params=params)
        for i in objs:
            self._tasks[action] = create_task_from_dict(i)
            return self._tasks[action]

    def _get_function_string(self) -> str:
        """
        Get function string.

        Returns
        -------
        str
            Function string.
        """
        return f"{self.kind}://{self.project}/{self.name}:{self.id}"

    #############################
    #  CRUD Methods for Tasks
    #############################

    def import_tasks(self, tasks: list[dict]) -> None:
        """
        Import tasks from yaml.

        Parameters
        ----------
        tasks : list[dict]
            List of tasks to import.

        Returns
        -------
        None
        """
        # Loop over tasks list, in the case where the function
        # is imported from local file.
        for task in tasks:
            # If task is not a dictionary, skip it
            if not isinstance(task, dict):
                continue

            # Create the object instance from dictionary,
            # the form in which tasks are stored in function
            # status
            task_obj = create_task_from_dict(task)

            # Try to save it in backend to been able to use
            # it for launching runs. In fact, tasks must be
            # persisted in backend to be able to launch runs.
            # Ignore if task already exists
            try:
                task_obj.save()
            except BackendError:
                pass

            # Set task if function is the same. Overwrite
            # status task dict with the new task object
            if task_obj.spec.function == self._get_function_string():
                self._tasks[task_obj.kind] = task_obj

    def new_task(self, **kwargs) -> Task:
        """
        Create new task.

        Parameters
        ----------
        **kwargs
            Keyword arguments.

        Returns
        -------
        Task
            New task.
        """
        kwargs["project"] = self.project
        kwargs["function"] = self._get_function_string()
        kwargs["kind"] = kwargs["kind"]
        task = new_task(**kwargs)
        self._tasks[kwargs["kind"]] = task
        return task

    def update_task(self, kind: str, **kwargs) -> None:
        """
        Update task.

        Parameters
        ----------
        kind : str
            Kind of the task.
        **kwargs
            Keyword arguments.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        self._raise_if_not_exists(kind)

        # Update kwargs
        kwargs["project"] = self.project
        kwargs["kind"] = kind
        kwargs["function"] = self._get_function_string()
        kwargs["uuid"] = self._tasks[kind].id

        # Update task
        task = create_task(**kwargs)
        task.save(update=True)
        self._tasks[kind] = task

    def get_task(self, kind: str) -> Task:
        """
        Get task.

        Parameters
        ----------
        kind : str
            Kind of the task.

        Returns
        -------
        Task
            Task.

        Raises
        ------
        EntityError
            If task is not created.
        """
        self._raise_if_not_exists(kind)
        return self._tasks[kind]

    def delete_task(self, kind: str, cascade: bool = True) -> None:
        """
        Delete task.

        Parameters
        ----------
        kind : str
            Kind of the task.
        cascade : bool
            Flag to determine if cascade deletion must be performed.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task is not created.
        """
        self._raise_if_not_exists(kind)
        delete_task(self.project, self._tasks[kind].name, cascade=cascade)
        self._tasks.pop(kind, None)

    def _raise_if_not_exists(self, kind: str) -> None:
        """
        Raise error if task is not created.

        Parameters
        ----------
        kind : str
            Kind of the task.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If task does not exist.
        """
        if self._tasks.get(kind) is None:
            raise EntityError("Task does not exist.")

    #############################
    #  Static interface methods
    #############################

    @staticmethod
    def _parse_dict(
        obj: dict,
        validate: bool = True,
    ) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        obj : dict
            Dictionary to parse.
        validate : bool
            Flag to determine if validation must be performed.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        project = obj.get("project")
        name = obj.get("name")
        kind = obj.get("kind")
        uuid = build_uuid(obj.get("id"))
        metadata = build_metadata(kind, framework_runtime=kind, **obj.get("metadata", {}))
        spec = build_spec(kind, framework_runtime=kind, validate=validate, **obj.get("spec", {}))
        status = build_status(kind, framework_runtime=kind, **obj.get("status", {}))
        return {
            "project": project,
            "name": name,
            "uuid": uuid,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
        }


def function_from_parameters(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    git_source: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = True,
    **kwargs,
) -> Function:
    """
    Create a new Function instance and persist it to the backend.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Name that identifies the object.
    kind : str
        Kind of the object.
    uuid : str
        ID of the object in form of UUID.
    description : str
        Description of the object.
    git_source : str
        Remote git source for object.
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object must be embedded in project.
    **kwargs
        Spec keyword arguments.

    Returns
    -------
    Function
       Object instance.
    """
    uuid = build_uuid(uuid)
    spec = build_spec(
        kind,
        framework_runtime=kind,
        **kwargs,
    )
    metadata = build_metadata(
        kind,
        framework_runtime=kind,
        project=project,
        name=name,
        version=uuid,
        description=description,
        source=git_source,
        labels=labels,
        embedded=embedded,
    )
    status = build_status(
        kind,
        framework_runtime=kind,
    )
    return Function(
        project=project,
        name=name,
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def function_from_dict(obj: dict) -> Function:
    """
    Create function from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create object from.

    Returns
    -------
    Function
        Function object.
    """
    return Function.from_dict(obj, validate=False)



from __future__ import annotations


class RuntimeRegistry:
    """
    Registry for runtimes to organize the import of classes.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        self.module = None
        self.class_name = None

    def register(self, module: str, class_name: str) -> None:
        """
        Register a runtime class.

        Parameters
        ----------
        module : str
            Module name.
        class_name : str
            Class name.

        Returns
        -------
        None
        """
        self.module = module
        self.class_name = class_name



"""
Task specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec, SpecParams


class TaskSpec(Spec):
    """Task specification."""

    def __init__(self, function: str) -> None:
        """
        Constructor.
        """
        self.function = function


class TaskParams(SpecParams):
    """
    Base task model.
    """

    function: str
    """Function string."""



"""
Context module.
"""
from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.projects.entity import Project


class Context:
    """
    Context class.

    It contains the project name, client and information about the type of client.
    It exposes CRUD operations for the entities and act as a layer between the
    project object and the client.
    The context is created by the context builder.
    """

    def __init__(self, project: Project) -> None:
        """
        Constructor.

        Parameters
        ----------
        project : Project
            The project object to create the context from.
        """
        self.name = project.name
        self.client = project._client
        self.local = project._client.is_local()
        self.project_dir = project.spec.context

    def create_object(self, api: str, obj: dict | None = None, **kwargs) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        return self.client.create_object(api, obj, **kwargs)

    def read_object(self, api: str, **kwargs) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict
            The read object.
        """
        return self.client.read_object(api, **kwargs)

    def update_object(self, api: str, obj: dict | None = None, **kwargs) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        return self.client.update_object(api, obj, **kwargs)

    def delete_object(self, api: str, **kwargs) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            The deleted object.
        """
        return self.client.delete_object(api, **kwargs)

    def list_objects(self, api: str, **kwargs) -> dict:
        """
        List objects.

        Parameters
        ----------
        api : str
            The api to list the objects with.
        filters : dict
            The filters to list the objects with.

        Returns
        -------
        dict
            The list of objects.
        """
        return self.client.list_objects(api, **kwargs)



"""
Spec factory entity.
"""
from __future__ import annotations

import typing

from digitalhub_core.utils.import_utils import (
    check_layer_existence,
    check_module_existence_by_framework,
    import_class,
    import_registry,
)

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.spec import Spec, SpecParams


def build_spec(
    kind: str,
    layer_digitalhub: str | None = None,
    framework_runtime: str | None = None,
    validate: bool = True,
    **kwargs,
) -> Spec:
    """
    Build entity spec object.

    Parameters
    ----------
    kind : str
        The type of spec to build.
    layer_digitalhub : str
        Layer name (e.g. "digitalhub_core") where the spec is defined.
    framework_runtime : str
        Framework name (e.g. "mlrun", "nefertem") where the spec is defined.
        This referes to the module where the spec is defined. The import
        function will search in different layers.
    validate : bool
        Flag to determine if arguments validation against a pydantic schema must be ignored.

    Returns
    -------
    Spec
        Spec object.
    """
    # Raise error if both layer and framework are not provided or if both are provided.
    # This is because we need some information where to search for the spec class.
    if (layer_digitalhub is None) == (framework_runtime is None):
        raise ValueError("Either layer or framework must be provided, but not both.")

    # If layer is not provided, try to import from runtime, otherwise search by layer.
    if framework_runtime is not None:
        layer_digitalhub = check_module_existence_by_framework(framework_runtime)
    else:
        check_layer_existence(layer_digitalhub)

    # Import registry
    # Note that this requires the creation of a registry whenever a new spec entity is created.
    module_to_import = f"{layer_digitalhub}.entities.registries"
    registry = import_registry(module_to_import, "spec_registry")

    # Import classes
    class_spec, class_params = _import_classes(registry, kind)

    # Validate arguments
    if validate:
        kwargs = class_params(**kwargs).dict(by_alias=True, exclude_none=True)

    return class_spec(**kwargs)


def _import_classes(registry: dict[str, dict[str, str]], kind: str) -> tuple[Spec, SpecParams]:
    """
    Import spec and params classes from registry.

    Parameters
    ----------
    registry : dict
        Registry.
    kind : str
        Entity kind.

    Returns
    -------
    tuple
        Spec class and params class.
    """
    try:
        module_and_class = registry[kind]
    except KeyError:
        raise ValueError(f"Spec {kind} not found in registry.")

    try:
        module = module_and_class["module"]
        spec_class = module_and_class["spec_class"]
        param_class = module_and_class["spec_params"]
    except KeyError:
        raise ValueError("Malformed registration.")

    return import_class(module, spec_class), import_class(module, param_class)



"""
Base Runtime module.
"""
from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable

from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.logger import LOGGER


class Runtime:
    """
    Base Runtime class.

    Runtimes are the entities responsible for the actual execution
    of a given run. They are highly specialized components which
    can translate the representation of a given execution as
    expressed in the run into an actual execution operation performed
    via libraries, code, external tools etc.
    Runtime types match function types.
    """

    ##################################
    # Abstract methods
    ##################################

    # This attribute is a list of allowed actions (tasks)
    # MUST BE explicitly extended in the subclass.
    allowed_actions = []

    def __init__(self) -> None:
        """
        Constructor.
        """
        if not self.allowed_actions:
            raise EntityError(
                "'allowed_actions' attribute must be extended in the subclass with the list of allowed actions."
            )

    @abstractmethod
    def build(self, function: dict, task: dict, run: dict) -> dict:
        """
        Build run spec.
        """

    @abstractmethod
    def run(self, run: dict) -> dict:
        """
        Execute run task.
        """

    @staticmethod
    @abstractmethod
    def _get_executable(action: str) -> Callable:
        """
        Get function from action.
        """

    ##################################
    # Private methods
    ##################################

    def _validate_task(self, run: dict) -> str:
        """
        Check if task is allowed. This presumes that the
        runtime holds a list of allowed actions in the self.allowed_actions
        attribute.

        Parameters
        ----------
        run : dict
            Run object dictionary.

        Returns
        -------
        str
            Action to execute.
        """
        action = self._get_action(run)
        if action not in self.allowed_actions:
            msg = f"Task {action} not allowed for {self.__class__.__name__} runtime."
            LOGGER.error(msg)
            raise EntityError(msg)
        return action

    @staticmethod
    def _get_action(run: dict) -> str:
        """
        Get action to execute from run.

        The run object stores in its spec the identifier of the task
        it is associated with. The task string is derived from the
        function string, and has the following format:

        <function-kind>+<task-action>://<function-name>:<function-id>

        Parameters
        ----------
        run : dict
            Run object dictionary.

        Returns
        -------
        str
            Action to execute.

        Raises
        ------
        RuntimeError
            If malformed run spec.

        Examples
        --------
        >>> run = {"spec": {"task": dbt+transform://dbt-example-function:some-uuid4}}
        >>> Runtime._get_action(run)
        'transform'
        """
        try:
            return run["spec"]["task"].split(":")[0].split("+")[1]
        except (KeyError, IndexError):
            msg = "Malformed run spec."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

    @staticmethod
    def _execute(func: Callable, *args, **kwargs) -> Any:
        """
        Execute function.

        Parameters
        ----------
        func : Callable
            Function to execute.
        *args
            Function arguments.
        **kwargs
            Function keyword arguments.

        Returns
        -------
        Any
            Function result.
        """
        try:
            return func(*args, **kwargs)
        except Exception:
            msg = "Something got wrong during function execution."
            LOGGER.exception(msg)
            raise RuntimeError(msg)



"""
DHCore Client module.
"""
from __future__ import annotations

import os

import requests
from digitalhub_core.client.objects.base import Client
from digitalhub_core.utils.exceptions import BackendError
from pydantic import BaseModel


class AuthConfig(BaseModel):
    """Client configuration model."""

    user: str = None
    """Username."""


class OAuth2TokenAuth(AuthConfig):
    """OAuth2 token authentication model."""

    access_token: str
    """OAuth2 token."""


class BasicAuth(AuthConfig):
    """Basic authentication model."""

    password: str
    """Basic authentication password."""


class ClientDHCore(Client):
    """
    DHCore client.

    The DHCore client is used to communicate with the Digitalhub Core backendAPI via REST.
    At creation, the client trys to get the endpoint and authentication parameters
    from the environment variables. In case the endpoint is not set, it raises an exception.
    """

    def __init__(self, config: dict = None) -> None:
        """
        Constructor.
        """
        super().__init__()

        self._endpoint = None

        self._auth_type = None
        self._user = None
        self._password = None
        self._access_token = None

        self._configure(config)

    def create_object(self, api: str, obj: dict | None = None, **kwargs) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        kwargs["json"] = obj
        return self._call("POST", api, **kwargs)

    def read_object(self, api: str, **kwargs) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict
            The object.
        """
        return self._call("GET", api, **kwargs)

    def update_object(self, api: str, obj: dict | None = None, **kwargs) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        kwargs["json"] = obj
        return self._call("PUT", api, **kwargs)

    def delete_object(self, api: str, **kwargs) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            A generic dictionary.
        """
        resp = self._call("DELETE", api, **kwargs)
        if isinstance(resp, bool):
            resp = {"deleted": resp}
        return resp

    def list_objects(self, api: str, **kwargs) -> list[dict]:
        """
        List objects.

        Parameters
        ----------
        api : str
            The api to list the objects with.
        **kwargs : dict


        Returns
        -------
        list[dict]
            The list of objects.
        """
        if kwargs is None:
            kwargs = {}

        if kwargs.get("params") is None:
            kwargs["params"] = {}

        start_page = 0
        if "page" not in kwargs["params"]:
            kwargs["params"]["page"] = start_page

        objects = []
        while True:
            resp = self._call("GET", api, **kwargs)
            contents = resp["content"]
            total_pages = resp["totalPages"]
            if not contents or kwargs["params"]["page"] >= total_pages:
                break
            objects.extend(contents)
            kwargs["params"]["page"] += 1

        return objects

    def _call(self, call_type: str, api: str, **kwargs) -> dict:
        """
        Make a call to the DHCore API.
        Keyword arguments are passed to the session.request function.

        Parameters
        ----------
        call_type : str
            The type of call to make.
        api : str
            The api to call.
        **kwargs
            Keyword arguments.

        Returns
        -------
        dict
            The response object.
        """
        url = self._endpoint + api

        # Choose auth type
        if self._auth_type == "basic":
            kwargs["auth"] = self._user, self._password
        elif self._auth_type == "oauth2":
            kwargs["headers"] = {"Authorization": f"Bearer {self._access_token}"}

        # Call
        response = None
        try:
            response = requests.request(call_type, url, timeout=60, **kwargs)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            if isinstance(e, requests.exceptions.Timeout):
                msg = "Request to DHCore backend timed out."
            elif isinstance(e, requests.exceptions.ConnectionError):
                msg = "Unable to connect to DHCore backend."
            elif isinstance(e, requests.exceptions.JSONDecodeError):
                return {}
            else:
                msg = f"Backend error: {e}"
            raise BackendError(msg) from e

    ################################
    # Configuration methods
    ################################

    def _configure(self, config: dict | None = None) -> None:
        """
        Function to set environment variables for DHub Core config.

        Parameters
        ----------
        config : ClientConfig
            The client config.

        Returns
        -------
        None
        """

        # Get endpoint at the beginning
        self._get_endpoint_from_env()

        # Evaluate configuration authentication parameters
        if config is not None:
            # Validate configuration against pydantic model

            # Try to get user/access_token or user/password

            if config.get("access_token") is not None:
                config = OAuth2TokenAuth(**config)
                self._user = config.user
                self._access_token = config.access_token
                self._auth_type = "oauth2"

            elif config.get("user") is not None and config.get("password") is not None:
                config = BasicAuth(**config)
                self._user = config.user
                self._password = config.password
                self._auth_type = "basic"

            return

        # Otherwise, use environment variables
        self._get_auth_from_env()

    def _get_endpoint_from_env(self) -> None:
        """
        Get DHub Core endpoint environment variables.

        Returns
        -------
        None

        Raises
        ------
        Exception
            If the endpoint of DHCore is not set in the env variables.
        """
        endpoint = os.getenv("DIGITALHUB_CORE_ENDPOINT")
        if endpoint is None:
            raise BackendError("Endpoint not set as environment variables.")

        # Sanitize endpoint string
        sanitized_endpoint = endpoint.removesuffix("/")

        # Set endpoint
        self._endpoint = sanitized_endpoint

    def _get_auth_from_env(self) -> None:
        """
        Get authentication parameters from the env.

        Returns
        -------
        tuple[str, str], str, None
            The authentication parameters.
        """
        # User for future entity ownership
        self._user = os.getenv("DIGITALHUB_CORE_USER")

        # Prioritize token over user/password
        token = os.getenv("DIGITALHUB_CORE_TOKEN")
        if token is not None:
            self._auth_type = "token"
            self._access_token = token
            return

        password = os.getenv("DIGITALHUB_CORE_PASSWORD")
        if self._user is not None and password is not None:
            self._auth_type = "basic"
            self._password = password
            return

    @staticmethod
    def is_local() -> bool:
        """
        Declare if Client is local.

        Returns
        -------
        bool
            False
        """
        return False
